# This file is part of cbp.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""computeHolePositions function: compute hole positions for a CBP mask."""

__all__ = ["computeHolePositions"]

import math

from lsst.geom import Point2D
from lsst.afw.cameraGeom import PIXELS, FIELD_ANGLE
from .coordUtils import fieldAngleToVector, vectorToFieldAngle


def computeHolePositions(detectorNames, detectorPositions, cameraGeom, cbpFlipX, cbpFocalLength):
    """Compute hole positions for a CBP mask.

    Given the desired locations of one or more spots on each detector,
    and assuming the telescope and CBP are pointing directly at each other,
    compute hole positions for a CBP mask.

    Parameters
    ----------
    detectorNames : `iterable` of `str`, or None,
        List of detector names; if None, use all detectors in ``cameraGeom``,
        sorted by name.
    detectorPositions : `iterable` of pair of `float`
        Detector x, y positions (pixels).
        Note that the center of a 1000x1000 pixel detector is (499.5, 499.5)
    cameraGeom : `lsst.afw.cameraGeom.Camera`
        Camera geometry.
    cbpFlipX : `bool`
        Is the CBP focal plane flipped?

    Returns
    -------
    holePositions : `list` of `tuple` of pair of `float`
        CBP mask x, y hole positions mm.

    Notes
    -----
    This code assumes that all detectors have approximately the same
    dimensions and orientation. This restriction should suffice for LSST
    because the two kinds of CCDs it uses have very similar dimensions.
    However, it will not do for HSC because that has very rectangular
    CCDs and some are 90 degrees from the others.
    """
    holePositions = []
    pixelPosList = [Point2D(*val) for val in detectorPositions]
    if detectorNames is None:
        detectorNames = sorted(list(cameraGeom.getNameIter()))
    for detectorName in detectorNames:
        detector = cameraGeom[detectorName]
        pixelSys = detector.makeCameraSys(PIXELS)
        pixelsToFieldAngle = cameraGeom.getTransform(pixelSys, FIELD_ANGLE)
        telFieldAngleList = pixelsToFieldAngle.applyForward(pixelPosList)
        for telFieldAngle in telFieldAngleList:
            # Compute hole positions for when the telescope and CBP
            # are pointing at each other;
            # thus CBP pupil vector = telescope pupil vector with z negated.
            # This is simply a shortcut for setting telAzAlt and cbpAzAlt,
            # then transforming vectors from tel pupil to base,
            # then from base to CBP pupil.
            telVector = fieldAngleToVector(telFieldAngle, False)
            cbpVector = (telVector[0], telVector[1], -telVector[2])
            cbpFieldAngle = vectorToFieldAngle(cbpVector, cbpFlipX)
            holePos = tuple(math.tan(ang) * cbpFocalLength for ang in cbpFieldAngle)
            holePositions.append(holePos)
    return holePositions
